/*
 inetrd.c Internet redirection super server
 Copyright  2002 by Padraig Brady <Padraig@antefacto.com>.

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include <stdarg.h>

#include "parsecfg.h"

#define VERSION "0.0.3" /* will be sent to HTTP clients */
#define MAX_REDIRECT_SERVERS 1000
#define MAX_REDIRECT_CLIENTS 300 /* Simultaneously connected */
#define NUM_CONFIGS 2
#define MAX_HOST_LEN 128

#if MAX_REDIRECT_CLIENTS > FD_SETSIZE
#   undef MAX_REDIRECT_CLIENTS
#   define MAX_REDIRECT_CLIENTS FD_SETSIZE
#endif

void DebugMsg(const char *fmt,...);
int writen(register int fd, register char* ptr, register int nbytes);
int setNonBlocking(int s);
int parseIP_PORT(const char* IP_PORT, unsigned long* ip, unsigned short* port);
void setConfigSignal(void);
void initClients(void);
void initServers(int currConfig);
int createServers(void);
void kickClient(int client_id, int currConfig);
void handleNewClients(time_t now, fd_set* rfds);
void handleExistingClients(time_t now, fd_set* rfds);
void handleExistingClient(time_t now, int client_id);

char* config_file = "/etc/inetrd.ini";
char* pidfile = NULL;
int client_timeout = 5; /* clients have 5 seconds to send valid request */
int interactive=0;

int getNewConfig=1;

/* If there is no parameter in the configuration file,
   the value is initialized 0 or NULL (boolean value is
   initialized -1(FALSE)). */

int currConfig=0;
int numRedirects[2]={0,0};
char** mode[2]={0,0};
char** uri[2]={0,0};
char** text[2]={0,0};

typedef enum {
    INVALID_MODE,
    WEB_ERROR,
    WEB_REDIR,
    BANNER_text,
    BANNER_file,
    TCP_REDIR,
    UDP_REDIR
} inetrd_mode_t;

/* TODO: Need to break one-one relationship between serverFDs and cfgini arrays.
         Instead to support virtual hosts serverFDs should be like:
         struct server {
             int fd;
             unsigned long ip;
             unsigned short port;
             int default_server_id; // points to id of ip:port or first member if no non vhost sections
             char* virtualhosts; // name1=server_id1,name2=server_id1,...
         } serverFDs[2][MAX_REDIRECT_SERVERS];
*/

struct server {
    int fd;
    unsigned long ip;
    unsigned short port;
    inetrd_mode_t mode; /* cache mode here so that don't compare strings each time */
} serverFDs[2][MAX_REDIRECT_SERVERS];

struct client {
    int fd;
    int server_id;
    time_t connect_time;
    char requested_host[MAX_HOST_LEN];
    int currConfig;
} clientFDs[MAX_REDIRECT_CLIENTS];
unsigned int currClients[2]={0,0};

cfgStruct cfgini[2][4] = {
    /* parameter  type  address of variable */
    {
    {"mode",    CFG_STRING, &mode[0]},
    {"WEB_REDIR_uri",    CFG_STRING, &uri[0]},
    {"WEB_ERROR_text",    CFG_STRING, &text[0]},
    {NULL, CFG_END, NULL}
    },{
    {"mode",    CFG_STRING, &mode[1]},
    {"WEB_REDIR_uri",    CFG_STRING, &uri[1]},
    {"WEB_ERROR_text",    CFG_STRING, &text[1]},
    {NULL, CFG_END, NULL}
    }
};

static void cfgErrorPrint(cfgErrorCode error_code, const char *file, int line, const char *str) {
    DebugMsg("Error at line %d of file %s [%s]", line, file, cfgStrError(error_code));
}

void newConfig(int signal) {
    /* init flag to get new config */
    getNewConfig=1;

    /* Note Glibc2 blocks any SIGHUPs while this is running,
       and it also resets the handler to this again, when
       finished, but doing explicitly for portability. */
    setConfigSignal();
}

void setConfigSignal(void) {
    signal(SIGHUP, newConfig);
}

void checkForNewConfig(void) {
    if (getNewConfig) {
        int newConfigError=createServers();
        if (newConfigError == -2) {/* very bad */
            DebugMsg("very bad config error, exiting");
            if (interactive) {
                exit(EXIT_FAILURE);
            }
        } else if (newConfigError == -1) {/* config error */
            currConfig ^= 1; //toggle back config
            if (!numRedirects[currConfig]) {/* It's not worth hanging around */
                DebugMsg("No current config, exiting");
                if (interactive) {
                    exit(EXIT_FAILURE);
                }
            }
            DebugMsg("Warning: New config ignored");
        }
        getNewConfig=0;
    }
}

void parseOptions(int argc, char** argv) {
    int index;
    int c;

    while ((c = getopt(argc, argv, "c:hip:t:")) != -1) {
        switch (c) {
            case 'c':
                config_file=optarg;
                break;
            case 'h':
                printf("%s [-c file] [-d level] [-h] [-i] [-t timeout]\n", argv[0]);
                printf("-c file, specifies the location of the config file. %s in the current directory is the default\n",
                       config_file);
                printf("-h prints this help\n");
                printf("-i turns on interactive mode, where %s runs in foreground\n", argv[0]);
                printf("-p pidfile, sets the location of the pidfile. None is written by default\n");
                printf("-t timeout, sets the time in seconds clients are given to send a valid HTTP request. "
                       "%ds is the default\n", client_timeout);
                exit(EXIT_SUCCESS);
                break;
            case 'i':
                interactive = 1;
                break;
            case 'p':
                pidfile=optarg;
                break;
            case 't':
                client_timeout = atoi(optarg);
                break;
            default:
                printf("Ignoring unknown option `-%c'.\n", optopt);
                break;
        }
    }

    for (index = optind; index < argc; index++)
        printf("Ignoring non-option argument %s\n", argv[index]);

    if (!interactive) {
        if (daemon(0, 0) == -1) {
            DebugMsg("daemon error [%d (%s)]", errno, strerror(errno));
            exit(EXIT_FAILURE);
        }
    }

    /* get our pid after (possible) daemon() above, because it could fork */
    if (pidfile != NULL) {
        /* Write the PID file. */
        FILE* pidfp = fopen(pidfile, "w");
        if (pidfp == (FILE*) 0) {
            DebugMsg("Could open [%s] for writing [%d (%s)]", pidfile, errno, strerror(errno));
            exit(EXIT_FAILURE);
        }
        (void) fprintf(pidfp, "%d\n", (int) getpid());
        (void) fclose(pidfp);
    }
}

int main(int argc, char *argv[]) {
    int currRedirect; int currClient;
    int configID;

    parseOptions(argc, argv);

    for (configID=0; configID<NUM_CONFIGS; configID++) {
        initServers(configID);
    }
    initClients();

    setConfigSignal();

    cfgSetFatalFunc(cfgErrorPrint);

    for (;;) {
        fd_set rfds;
        time_t now;
        int max_fd=0;
        struct timeval tv;
        int numWaiting;

        checkForNewConfig();

        FD_ZERO(&rfds);

        /* check for new clients */
        for (currRedirect=0; currRedirect<numRedirects[currConfig]; currRedirect++) {
            FD_SET(serverFDs[currConfig][currRedirect].fd, &rfds);
            if (serverFDs[currConfig][currRedirect].fd > max_fd)
                max_fd = serverFDs[currConfig][currRedirect].fd;
        }
        /* check for request from existing clients */
        for (currClient=0; currClient<MAX_REDIRECT_CLIENTS; currClient++) {
            if (clientFDs[currClient].fd != -1) {
                FD_SET(clientFDs[currClient].fd, &rfds);
                if (clientFDs[currClient].fd > max_fd)
                    max_fd = clientFDs[currClient].fd;
            }
        }

        tv.tv_sec = 2; /* seconds resolution to kick "bad" clients off */
        tv.tv_usec = 0;

        /* Note I don't have to worry if a signal arrives between checkForNewConfig()
           and here, as select will never block indefinitely. I.E. I don't need to use pselect() */
        numWaiting = select(max_fd+1, &rfds, (fd_set*) NULL, (fd_set*) NULL, &tv);
        now=time(0);
        if (numWaiting < 0) {
            if (errno!=EINTR) {
                DebugMsg("select error [%d (%s)]", errno, strerror(errno));
                return EXIT_FAILURE; //assume this error isn't transient?
            }
        } else if (numWaiting == 0) { //after timeout
            handleExistingClients(now, &rfds); /* (see if we can kick any off) */
        } else {
            handleExistingClients(now, &rfds);
            handleNewClients(now, &rfds);
        }
    }
}

void initClients(void) {
    int currClient;
    for (currClient=0; currClient<MAX_REDIRECT_CLIENTS; currClient++)
        clientFDs[currClient].fd=-1;
}

void initServers(int currConfig) {
    int currRedirect;
    for (currRedirect=0; currRedirect<MAX_REDIRECT_SERVERS; currRedirect++) {
        serverFDs[currConfig][currRedirect].fd=-1;
        serverFDs[currConfig][currRedirect].ip=0;
        serverFDs[currConfig][currRedirect].port=0; /* port == 0 is always invalid */
        serverFDs[currConfig][currRedirect].mode=INVALID_MODE;
    }
}

/* Returns -1 on config error,
           -2 on fatal error,
            0 on success
*/
int createServers(void) {

    int currRedirect, oldCurrRedirect;
    int oldConfig = currConfig;
    currConfig ^= 1; //toggle config

    /* Kick any clients off current config as it's invalid */
    if (currClients[currConfig] != 0) {
        int currClient;

        DebugMsg("Warning: there were %d clients on the current config (they were kicked off)", currClients[currConfig]);

        for (currClient=0; currClient<MAX_REDIRECT_CLIENTS; currClient++) {
            if ((clientFDs[currClient].fd != -1) && (clientFDs[currClient].currConfig == currConfig)) {
                kickClient(currClient, currConfig);
            }
        }

        cfgFree(cfgini[currConfig], CFG_INI, numRedirects[currConfig]);
    }

    /* Read new config in */
    if ((numRedirects[currConfig] = cfgParse(config_file, cfgini[currConfig], CFG_INI)) == -1)
        return -1;
    if (numRedirects[currConfig] > MAX_REDIRECT_SERVERS) {
        DebugMsg("Too many servers specified, ignoring new config");
        return -1;
    }

    for (currRedirect=0; currRedirect<numRedirects[currConfig]; currRedirect++) {
        if (!parseIP_PORT(cfgSectionNumberToName(currRedirect),
                          &serverFDs[currConfig][currRedirect].ip,
                          &serverFDs[currConfig][currRedirect].port)) {
            DebugMsg("Invalid ip:port specified [%s]", cfgSectionNumberToName(currRedirect));
            return -1;
        }
    }

    for (currRedirect=0; currRedirect<numRedirects[currConfig]; currRedirect++) {

        struct sockaddr_in servAddr;
        int one=1;
        int sd=-1;

        unsigned long ip=serverFDs[currConfig][currRedirect].ip;
        unsigned short port=serverFDs[currConfig][currRedirect].port;

        /* If socket already in oldConfig then use it */
        for (oldCurrRedirect=0; oldCurrRedirect<numRedirects[oldConfig]; oldCurrRedirect++) {
            if ((ip == serverFDs[oldConfig][oldCurrRedirect].ip) && (port == serverFDs[oldConfig][oldCurrRedirect].port)) {
                sd = serverFDs[oldConfig][oldCurrRedirect].fd;
                serverFDs[oldConfig][oldCurrRedirect].fd = -1;
                break;
            }
        }

        if (sd == -1) {
            /* create socket */
            sd = socket(AF_INET, SOCK_STREAM, 0);
            if(sd<0) {
                DebugMsg("cannot open socket [%d (%s)]", errno, strerror(errno));
                return -2;
            }

            /* bind server port */
            servAddr.sin_family = AF_INET;
            servAddr.sin_port = port;
            servAddr.sin_addr.s_addr = ip;

            setsockopt(sd,SOL_SOCKET,SO_REUSEADDR,&one,sizeof(one));

            if(bind(sd, (struct sockaddr *) &servAddr, sizeof(servAddr))<0) {
                DebugMsg("cannot bind port [%d (%s)]", errno, strerror(errno));
                return -2;
            }

            if(listen(sd,SOMAXCONN)<0) {
                DebugMsg("listen failed [%d (%s)]", errno, strerror(errno));
                return -2;
            }

            /* This is done so that accept() will not block
             * in the presence of "asynchronous network errors"
             * between the select() and accept() calls. */
            if(setNonBlocking(sd)<0) {
                DebugMsg("setNonBlocking failed [%d (%s)]", errno, strerror(errno));
                return -2;
            }
        }

        serverFDs[currConfig][currRedirect].fd=sd;

        /* Cache ids here so that don't have to do strcmps each time */
        if (!strcmp(mode[currConfig][currRedirect],"WEB_ERROR"))
            serverFDs[currConfig][currRedirect].mode=WEB_ERROR;
        else if (!strcmp(mode[currConfig][currRedirect],"WEB_REDIR"))
            serverFDs[currConfig][currRedirect].mode=WEB_REDIR;
        else
            serverFDs[currConfig][currRedirect].mode=INVALID_MODE;
    }

    /* close any servers not needed */
    for (oldCurrRedirect=0; oldCurrRedirect<numRedirects[oldConfig]; oldCurrRedirect++) {
        if (serverFDs[oldConfig][oldCurrRedirect].fd != -1) {
            close(serverFDs[oldConfig][oldCurrRedirect].fd);
        }
    }

    initServers(oldConfig);

    if (currClients[oldConfig] == 0)
        cfgFree(cfgini[oldConfig], CFG_INI, numRedirects[oldConfig]);

    return 0;
}

void handleNewClients(time_t now, fd_set* rfds) {
    int currRedirect, currClient;

    if (currClients[currConfig] >= MAX_REDIRECT_CLIENTS)
        return; /* allow OS to queue rather than rejecting connection */

    for (currRedirect=0; currRedirect<numRedirects[currConfig]; currRedirect++) {
        if (FD_ISSET(serverFDs[currConfig][currRedirect].fd, rfds)) {
            int newSd;

            for (;;) {
                newSd = accept(serverFDs[currConfig][currRedirect].fd, (struct sockaddr *) NULL, 0);
                if(newSd<0) {
                    if (newSd == EINTR)
                        continue;
                    else
                        /* This will not happen if client disconnects between the select and accept.
                           In that case the accept will succeed and read will return 0. This just
                           catches (logs) async network errors that could have occurred.
                           Note accept would block in this case if we did not mark the fd as non-blocking.*/
                        DebugMsg("cannot accept connection [%d (%s)]", errno, strerror(errno));
                }
                break;
            }

            if (newSd > 0) {
                currClients[currConfig]++;
                for (currClient=0; currClient<MAX_REDIRECT_CLIENTS; currClient++) {
                    if (clientFDs[currClient].fd == -1) {
                        clientFDs[currClient].fd=newSd;
                        clientFDs[currClient].connect_time=now;
                        clientFDs[currClient].server_id=currRedirect;
                        clientFDs[currClient].requested_host[0]='\0';
                        clientFDs[currClient].currConfig=currConfig;
                        break;
                    }
                }
            }
        }
    }
}

void handleExistingClients(time_t now, fd_set* rfds) {
    int currClient;
    int oldConfig = currConfig ^ 1;

    for (currClient=0; currClient<MAX_REDIRECT_CLIENTS; currClient++) {
        if (clientFDs[currClient].fd != -1) {
            long connected_time = now - clientFDs[currClient].connect_time;
            if (connected_time<0) { //Time must have changed backwards
                DebugMsg("Warning: time change (backwards). Applying workaround");
                clientFDs[currClient].connect_time = now; //reset timer
            } else if (connected_time > client_timeout+3) { //Time must have changed forwards (3 second processing allowed)
                DebugMsg("Warning: time change (forwards). Applying workaround");
                clientFDs[currClient].connect_time = now; //reset timer
            } else if (connected_time >= client_timeout) {
                kickClient(currClient, clientFDs[currClient].currConfig);
            } else if (FD_ISSET(clientFDs[currClient].fd, rfds)) {
                handleExistingClient(now, currClient);
            }
        }
    }
    if (currClients[oldConfig] == 0)
        cfgFree(cfgini[oldConfig], CFG_INI, numRedirects[oldConfig]);
}

/* Note read/write should not block here as they'll return 0 if there's an error on the socket */
void handleExistingClient(time_t now, int client_id) {
    int  server_id;
    int  currConfig;
    int  client_response_size=0;
    char client_response[1024];
    char HTTP_Date[256];
    char client_request[2049];
    int  client_request_size;
    int  client_finished=0;
    char* host;
    inetrd_mode_t mode;

    server_id = clientFDs[client_id].server_id;
    currConfig = clientFDs[client_id].currConfig;
    host = clientFDs[client_id].requested_host;

    mode = serverFDs[currConfig][server_id].mode;

    client_request_size=read(clientFDs[client_id].fd, client_request, sizeof(client_request)-1);
    if (client_request_size < 1) {
        client_finished=1;
    } else {
        client_request[client_request_size]='\0'; /* Terminate string as usually text protocols */

        switch (mode) {
        case WEB_REDIR:
        case WEB_ERROR:

            if (host[0] == '\0') {
                /* Note host contains everything entered in browser
                   between type and /, i.e. the xxxx:xx in
                   http://xxxx:xx/..... i.e. the port is also sent if entered */
                char *host_str, *EOL;
                host_str = strstr(client_request, "\nHost: ");
                if (host_str) {
                    host_str += 7;
                    EOL=strchr(host_str,'\r');
                    if (EOL) {
                        int host_len = EOL-host_str;
                        if (host_len > MAX_HOST_LEN-1)
                            host_len = MAX_HOST_LEN-1;
                        memcpy(host, host_str, host_len);
                        host[host_len] = '\0';
                    }
                }
            }

            if (strstr(client_request,"\r\n\r\n")) { /* assume this comes together in buffer */
                client_finished=1;
                client_response_size=0;

                (void) strftime(HTTP_Date, sizeof(HTTP_Date), "%a, %d %b %Y %H:%M:%S GMT", gmtime(&now));

                switch (mode) {
                    case WEB_REDIR:

                        client_response_size = snprintf(client_response, sizeof(client_response),
                        "HTTP/1.1 302 Temporarily Unavailable\r\n"
                        "Server: inetrd "VERSION"\r\n"
                        "Date: %s\r\n"
                        "Expires: %s\r\n"
                        "Location: %s\r\n"
                        "Cache-Control: no-cache\r\n" /* Note: Most HTTP/1.0 caches will not recognize or obey this directive */
                        "Connection: close\r\n"
                        "\r\n",
                        HTTP_Date,HTTP_Date,uri[currConfig][server_id]);

                        break;

                    case WEB_ERROR: {

                        /* Note with "503 Service Unavailable" you can specify the field:
                           "Retry-After: {date,seconds}" to tell client not to retry until this */
                        char* error_text=text[currConfig][server_id];
                        if (!error_text)
                           error_text = "Sorry, this site is temporarily unavailable";
                        client_response_size = snprintf(client_response, sizeof(client_response),
                        "HTTP/1.0 500 Internal Error\n"
                        "Server: inetrd "VERSION"\r\n"
                        "Content-type: text/html; charset=iso-8859-1\r\n"
                        "Date: %s\r\n"
                        "Expires: %s\r\n"
                        "Cache-Control: no-cache\r\n" /* Note: Most HTTP/1.0 caches will not recognize or obey this directive */
                        "Connection: close\r\n"
                        "Content-length: %d\r\n"
                        "\r\n"
                        "<HTML><BODY>%s</BODY></HTML>",
                        HTTP_Date,HTTP_Date,strlen(error_text)+26,error_text);

                        break;
                    }

                    default:
                        break;
                }
                writen(clientFDs[client_id].fd, client_response, client_response_size);
            }
            break;
        default:

            /* Other modes not handled at present */
            break;
        }
    }

    if (client_finished)
        kickClient(client_id, currConfig);
}

void kickClient(int client_id, int currConfig) {
    /* linux will send any pending data before close
       even if SO_LINGER is not set. Double check this
       as I noticed (once) both mozilla & links giving "connection reset"
       errors when connecting to inetrd on localhost.
       I.E. is there any race between:
           server write
           server close
       Hmm, probably (in kernel or more likely FIN could reach client before data on NET?) :-( */
    /* Note we usually let the clients initiate the close so that
     * the TIME_WAIT state connections will be spread across all the clients
     * rather than all being on this server */
    if (close(clientFDs[client_id].fd)<0)
        DebugMsg("Close error [%d (%s)]", errno, strerror(errno));
    clientFDs[client_id].fd = -1;
    currClients[currConfig]--;
}

/* converts IP_PORT of format 123.123.123.123:1234
   and stores appropriate values in ip & port
   0 returned on error
   1 returned on success

   Hmm could probably simplify this function with:
   sscanf(IP_PORT,"%15[.0-9]s:%4[0-9]s, ip_str, port_str) */
int parseIP_PORT(const char* IP_PORT, unsigned long* ip, unsigned short* port) {
    char* _pos;
    int ip_len;
    char ip_str[16];
    int port_tmp;

    _pos = strchr(IP_PORT, ':');
    if (_pos) {
        ip_len = _pos - IP_PORT;
        if (ip_len > (sizeof(ip_str)-1))
            return 0;
        else {
            memcpy(ip_str, IP_PORT, ip_len);
            ip_str[ip_len]='\0';
            *ip = inet_addr(ip_str);
            if (*ip == -1)
                return 0;
        }

        port_tmp = atoi(_pos+1);
        if (port_tmp<1 || port_tmp>65535)
            return 0;
        else
            *port = htons(port_tmp);
    } else {
        return 0;
    }
    return 1;
}

/*
 * This is a "safe" version of write. The standard write
 * may return before all the requested bytes are written because
 * the internal buffer (4096 bytes for a pipe for e.g. in some UNIX')
 * has filled. Taken from Stevens' "Unix Network Programming".
 */
int writen(register int fd, register char* ptr, register int nbytes) {
    int nleft, nwritten;

    nleft = nbytes;
    while(nleft > 0) {
        nwritten = write(fd, ptr, nleft);
        if (nwritten < 0) {
            if (errno==EINTR) continue; /* If signal interupted us, do send again */
            return nwritten; /* the error */
        }

        nleft -= nwritten;
        ptr += nwritten;
    }
    return (nbytes - nleft); /* return >=0 */
}

/* return 0 on success, -1 on fail */
int setNonBlocking(int s) {
    int flags;
    if ((flags = fcntl(s, F_GETFL, 0)) < 0)
        return -1;

    flags |= O_NONBLOCK;

    if (fcntl(s, F_SETFL, flags) < 0)
        return -1;

    return 0;
}

/* Just call like printf. Don't put \n at end of message */
void DebugMsg(const char *fmt,...) {
    int n=0,size = 256;
    char *buff;
    va_list ap;

    buff = (char *)malloc(size);
    if(!buff) {
        return;
    }

    va_start(ap,fmt);

    n = vsnprintf(buff,size,fmt,ap);

    if(size <= n) {
        size = n+1; //As of glibc 2.1 (C99 standard) vsnprintf sets n to required size (not including NULL)
        buff = (char *)realloc(buff,size);
        if(!buff) {
            return;
        }
        vsnprintf(buff,size,fmt,ap);
    }

    va_end(ap);

    if (interactive)
        printf("%s\n", buff);
    else
        syslog(LOG_DAEMON+LOG_ERR,"%s", buff); /* Note syslog converts any \n to space */

    free(buff);
}
